/*
 * $Log:	arcunp.c,v $
 * Revision 1.1  88/04/11  18:42:39  hyc
 * Initial revision
 * 
 * Revision 1.3  87/12/20  03:10:14  hyc
 * Ah hah! The value fgetc was returning was getting sign extended!
 * Thus, 0xff was indistinguishable from EOF, causing the premature
 * termination... The solution is to use getc, which doesn't do this
 * (on the Apollos, anyway...)
 * 
 * Note that I could *not* duplicate this behavior in a 5 liner...?
 * 
 * Revision 1.2  87/12/19  04:11:22  hyc
 * Move image mode code around from MTS only to non-MSDOS.
 * 
 * Revision 1.1  87/12/19  04:06:39  hyc
 * Initial revision
 * 
 * Revision 1.3  87/08/13  17:06:08  hyc
 * Run thru indent, fixed some signed vs. unsigned problems
 * with bp <-> buf, and inbuf and localbuf...
 *  Revision 1.2  87/07/21  09:28:31  hyc *** empty
 * log message ***
 * 
 */

/*
 * ARC - Archive utility - ARCUNP
 * 
 * Version 3.17, created on 02/13/86 at 10:20:08
 * 
 * (C) COPYRIGHT 1985 by System Enhancement Associates; ALL RIGHTS RESERVED
 * 
 * By:  Thom Henderson
 * 
 * Description: This file contains the routines used to expand a file when
 * taking it out of an archive.
 * 
 * Language: Computer Innovations Optimizing C86
 */
#include <stdio.h>
#include "arc.h"

/* stuff for repeat unpacking */

#define DLE 0x90		/* repeat byte flag */

static INT      state;		/* repeat unpacking state */

/* repeat unpacking states */

#define NOHIST 0		/* no relevant history */
#define INREP 1			/* sending a repeated value */

static INT      crcval;		/* CRC check value */
static LONG     size;		/* bytes to read */

INT
unpack(f, t, hdr)		/* unpack an archive entry */
	FILE           *f, *t;	/* source, destination */
	struct heads   *hdr;	/* pointer to file header data */
{
	INT             c;	/* one char of stream */
	INT             putc_unp();
	INT             putc_ncr();
	INT             getc_unp();

	/* setups common to all methods */

	crcval = 0;		/* reset CRC check value */
	size = hdr->size;	/* set input byte counter */
	state = NOHIST;		/* initial repeat unpacking state */
	setcode();		/* set up for decoding */

	/* use whatever method is appropriate */

	switch (hdrver) {	/* choose proper unpack method */
	case 1:		/* standard packing */
	case 2:
		while ((c = getc_unp(f)) != EOF)
			putc_unp(c, t);
		break;

	case 3:		/* non-repeat packing */
		while ((c = getc_unp(f)) != EOF)
			putc_ncr(c, t);
		break;

	case 4:		/* Huffman squeezing */
		init_usq(f);
		while ((c = getc_usq(f)) != EOF)
			putc_ncr(c, t);
		break;

	case 5:		/* Lempel-Zev compression */
		init_ucr(0);
		while ((c = getc_ucr(f)) != EOF)
			putc_unp(c, t);
		break;

	case 6:		/* Lempel-Zev plus non-repeat */
		init_ucr(0);
		while ((c = getc_ucr(f)) != EOF)
			putc_ncr(c, t);
		break;

	case 7:		/* L-Z plus ncr with new hash */
		init_ucr(1);
		while ((c = getc_ucr(f)) != EOF)
			putc_ncr(c, t);
		break;

	case 8:		/* dynamic Lempel-Zev */
		decomp(f, t);
		break;

	case 9:		/* Squashing */
		sqdecomp(f, t);
		break;

	default:		/* unknown method */
		if (warn) {
			printf("I don't know how to unpack file %s\n", hdr->name);
			printf("I think you need a newer version of ARC\n");
			nerrs++;
		}
		fseek(f, hdr->size, 1);	/* skip over bad file */
		return 1;	/* note defective file */
	}

	/* cleanups common to all methods */

	if (crcval != hdr->crc) {
		if (warn || kludge) {
			printf("WARNING: File %s fails CRC check\n", hdr->name);
			nerrs++;
		}
		return 1;	/* note defective file */
	}
	return 0;		/* file is okay */
}

/*
 * This routine is used to put bytes in the output file.  It also performs
 * various housekeeping functions, such as maintaining the CRC check value.
 */

INT
putc_unp(c, t)			/* output an unpacked byte */
	char            c;	/* byte to output */
	FILE           *t;	/* file to output to */
{
	crcval = addcrc(crcval, c);	/* update the CRC check value */
#ifdef	MTS
	if (!image)
		atoe(&c, 1);
#endif
#ifndef MSDOS
	if ((c != '\r') || image)
#endif
		putc_tst(c, t);
}

/*
 * This routine is used to decode non-repeat compression.  Bytes are passed
 * one at a time in coded format, and are written out uncoded. The data is
 * stored normally, except that runs of more than two characters are
 * represented as:
 * 
 * <char> <DLE> <count>
 * 
 * With a special case that a count of zero indicates a DLE as data, not as a
 * repeat marker.
 */

INT
putc_ncr(c, t)			/* put NCR coded bytes */
	unsigned char   c;	/* next byte of stream */
	FILE           *t;	/* file to receive data */
{
	static INT      lastc;	/* last character seen */

	switch (state) {	/* action depends on our state */
	case NOHIST:		/* no previous history */
		if (c == DLE)	/* if starting a series */
			state = INREP;	/* then remember it next time */
		else
			putc_unp(lastc = c, t);	/* else nothing unusual */
		return;

	case INREP:		/* in a repeat */
		if (c)		/* if count is nonzero */
			while (--c)	/* then repeatedly ... */
				putc_unp(lastc, t);	/* ... output the byte */
		else
			putc_unp(DLE, t);	/* else output DLE as data */
		state = NOHIST;	/* back to no history */
		return;

	default:
		abort("Bad NCR unpacking state (%d)", state);
	}
}

/*
 * This routine provides low-level byte input from an archive.  This routine
 * MUST be used, as end-of-file is simulated at the end of the archive entry.
 */

INT
getc_unp(f)			/* get a byte from an archive */
	FILE           *f;	/* archive file to read */
{
	register INT    xx;
	if (!size)		/* if no data left */
		return EOF;	/* then pretend end of file */

	size--;			/* deduct from input counter */
	xx = getc(f);
	return code(xx);	/* and return next decoded byte */
}
